import os
import re

XENBUS = ["/var/run/xenstored/socket", "/proc/xen/xenbus"]

def get_xenbus():
    for f in XENBUS:
        if os.path.isfile(f):
            return f

    raise RuntimeError("Cannot find xenbus")

def xenbus_read(key):
    env_key = ("XENBUS_%s" % re.sub('[-/]', '_', key)).upper()

    # allow a local override for testing and odd situations
    if env_key in os.environ:
        return os.environ[env_key]

    # this is easier and faster then forking a subprocess, i promise
    fd = os.open(get_xenbus(), os.O_RDWR)
    length = chr(len(key) + 1)
    os.write(fd, "\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"+length+"\x00\x00\x00")
    os.write(fd, key+"\x00")

    junk = os.read(fd, 16)
    data = os.read(fd, 512)
    os.close(fd)

    return data

def instance_id():
    data = xenbus_read("name")
    m = re.search('instance-([a-z0-9-]+)$', data)

    if not m:
        raise RuntimeError("Cannot parse response from xenbus")
    else:
        return m.group(1)

def instance_name():
    return xenbus_read("vm-data/hostname")

def instance_region():
    return xenbus_read("vm-data/provider_data/region")




import os
import time
import logging
import pyrax
from pyrax.utils import wait_for_build
from paramiko import RSAKey

from serverondemand import get_resource
from serverondemand.xen import instance_id, instance_name
from serverondemand.executor import Executor

logger = logging.getLogger(__name__)

def servers_api(creds_file, region = None):
    """Sets up a pyrax instance for Cloud Servers in the given region."""
    pyrax.set_setting("identity_type", "rackspace")
    pyrax.set_credential_file(os.path.expanduser(creds_file), region = region)
    cs = pyrax.cloudservers

    return cs

class Demand(object):
    def __init__(self, args):
        self.debug = not not args.debug
        self.script = args.script if os.path.isfile(args.script) else False
        self.server = None

        region = args.region.upper() if args.region else None
        self.api = servers_api(args.credentials, region)

        if args.agent_creds:
            self.agent_username = args.agent_creds['username']
            self.agent_apikey = args.agent_creds['apikey']
        else:
            logger.debug("Loading agent credentials from %s" % args.credentials)
            from ConfigParser import SafeConfigParser
            parser = SafeConfigParser()
            parser.read(os.path.expanduser(args.credentials))
            self.agent_username = parser.get('rackspace_cloud', 'username')
            self.agent_apikey = parser.get('rackspace_cloud', 'api_key')

    def build(self, image, flavor, **kwargs):
        """Builds a server and runs the requested script on it.

        image and flavor are values that will be recognized by the novaclient library
        as references to an image and a flavor to use to boot the server with.

        Optionally, other kewword arguments can be passed and will be passed allong to
        the novaclient.boot method to enable other things such as more metadata, disk
        configuration options, network interfaces, etc...
        """
        logger.debug("Generating new SSH key")
        name = "serverondemand-%s" % str(time.time()).replace('.', '-')

        key = RSAKey.generate(2048)
        personality = {
                "/root/.ssh/authorized_keys": "ssh-rsa %s serverondemand generated key\n" % key.get_base64(),
                }
        meta = {
                'origin': 'serverondemand',
                }

        if kwargs.get('meta', None):
            meta.update(kwargs.pop('meta'))
        if kwargs.get('files', None):
            personality.update(kwargs.pop('files'))

        logger.info("Building server")

        server = self.api.servers.create(name, image, flavor, meta = meta, files = personality, **kwargs)
        server = wait_for_build(server, verbose = self.debug)

        self.server = server
        self.key = key

        logger.info("Uploading bootstrap")
        self.bootstrap()
        logger.info("Executing scripts")
        self.run_script()

        # all done

    def bootstrap(self):
        """Uploads scripts to the server.

        Creates the demand_done script which the new server can use to delete
        itself, along with the bootstrap.sh script that sets up useful things."""
        executor = Executor(self.server.accessIPv4, 'root', self.key)
        sftp = executor.open_sftp()

        self._write_sftp_file(sftp, '/usr/sbin/demand_done', mode=0755, data =
                "#!/usr/bin/env python\n" +
                "RS_USERNAME='%s'\n" % self.agent_username +
                "RS_APIKEY='%s'\n" % self.agent_apikey +
                get_resource('xen.py') +
                get_resource('resources/demand_done.py'))

        self._write_sftp_file(sftp, '/root/bootstrap.sh',
                get_resource('resources/bootstrap.sh'), 0755)

        if self.script:
            with open(self.script, 'r') as fd:
                self._write_sftp_file(sftp, '/root/script.sh', fd.read(), 0755)

        executor.close()

    def _write_sftp_file(self, sftp, path, data, mode = 0644):
        fd = sftp.open(path, 'w')
        fd.chmod(mode)
        written = fd.write(data)
        fd.close()
        return written

    def run_script(self):
        """Runs the scripts on the server.

        First runs the bootstrap.sh script we generated, and waits for it to complete.

        If the user provided a script of their own, it runs that too, but detatches
        and lets it run in the background."""
        executor = Executor(self.server.accessIPv4, 'root', self.key, debug = self.debug)

        res = executor.run("/root/bootstrap.sh")
        logger.debug("Bootstrap script: %s" % res)

        logger.info("Running user script.sh detatched.")
        if self.script:
            res = executor.run("nohup /root/script.sh")

        executor.close()

    def get_flavor(self, ref):
        """Finds the perfect server flavor

        Pass in a flavor ID and it will return the Flavor object with its name et. al.

        # TODO I'd love a nice way to lazy specify what type of flavor we want """
        #if 'GB' not in ref and 'MB' not in ref:
        flavor = self.api.flavors.get(ref)
        return flavor.name, flavor.id

        #potential = [ flv for flv in self.api.flavors.list()
        #        if substr in flv.name.replace(' ', '') ]
        #logger.debug("Potential Flavors: %s" % [flv.name for flv in potential])

        #potential.sort(cmp = lambda a, b: 1 if a.name > b.name else -1)
        #flavor = potential[-1]

        #return flavor.name, flavor.id

    def get_image(self, substr):
        """Finds the perfect server image

        Pass in a random string and it will find the "highest" image that
        contains that string in its name.  Useful for getting the most recent
        "Ubuntu" image or something like that."""
        potential = [ img for img in self.api.images.list()
                if substr.lower() in img.name.lower() ]
        logger.debug("Potential Images: %s" % [img.name for img in potential])

        potential.sort(cmp = lambda a, b: 1 if a.name > b.name else -1)
        image = potential[-1]

        return image.name, image.id




import re
import logging
from socket import timeout as sock_timeout

from paramiko import SSHClient
from paramiko.client import MissingHostKeyPolicy

logger = logging.getLogger(__name__)

class Executor(object):
    def __init__(self, hostname, username, key, debug=False):
        self.debug = debug
        self.client = SSHClient()
        self.client.set_missing_host_key_policy(MissingHostKeyPolicy())
        self.client.connect(hostname, username=username, pkey=key,
                look_for_keys=False, compress=True)

        self.chan = self.client.invoke_shell()
        self.chan.settimeout(0.1)
        self.prompt_regex = "# $"
        self.read()

        self.set_prompt()

    def set_prompt(self):
        prompt = "fio8g3rfubsd97g34iu:# "
        self.chan.send('export PS1="%s" PS2=""\n' % prompt)
        self.prompt_regex = re.escape(prompt) + '$'
        self.read()

    def run(self, cmdline):
        """Run a command and piece out the console output."""
        # run the command
        cmdline += "; echo\n"
        self.chan.send(cmdline)

        # piece out the response
        response = self.read()
        m = re.match(re.escape(cmdline) + "(.*)" + self.prompt_regex, response)

        # always return something
        if m:
            return m.group(1)
        else:
            return response

    def recv(self, how_much = 1024):
        """Reads bytes from the session, catching timouts."""
        try:
            data = self.chan.recv(how_much)
            logger.debug("recv %d bytes: %s" % (len(data), repr(data)))
            return data
        except sock_timeout:
            return None

    def read(self):
        """Reads until the prompt is seen again."""
        data = ''
        buf = self.recv()

        while not re.search(self.prompt_regex, data):
            if buf is not None:
                data += buf
            buf = self.recv()

        logger.debug("returning buffer: %s" % data)
        return data

    def open_sftp(self):
        return self.client.open_sftp()

    def close(self):
        try: self.chan.close()
        except: pass

        try: self.client.close()
        except: pass




from setuptools import setup, find_packages

setup(name='serverondemand',
    version='0.0.1',
    author='Andrew Regner',
    author_email='andrew@aregner.com',
    description='Wrapper library / script to run things on a as-needed server.',
    packages=['serverondemand'],
    scripts=[
        "bin/demand_server",
        ],
    package_data={
        '': ['*.sh', '*.py'],
        },
    install_requires=[
        'pyrax',
        'paramiko'
        ])

# vim:et:fdm=marker:sts=4:sw=4:ts=4
